from django.apps import apps as django_apps
from django.conf import settings
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from ..creators import UnscheduledAppointmentCreator


class Helper:
    def __init__(self, subject_identifier=None, now=None):
        self.subject_identifier = subject_identifier
        self.now = now or get_utcnow()

    @property
    def consent_model_cls(self):
        """Returns a consent model class.

        Defaults to edc_appointment.subjectconsent
        """
        try:
            return django_apps.get_model(settings.SUBJECT_CONSENT_MODEL)
        except LookupError:
            return django_apps.get_model("edc_appointment.subjectconsent")

    def consent_and_put_on_schedule(
        self,
        subject_identifier=None,
        visit_schedule_name=None,
        schedule_name=None,
    ):
        subject_identifier = subject_identifier or self.subject_identifier
        subject_consent = self.consent_model_cls.objects.create(
            subject_identifier=subject_identifier, consent_datetime=self.now
        )
        visit_schedule = site_visit_schedules.get_visit_schedule(
            visit_schedule_name or "visit_schedule1"
        )
        schedule = visit_schedule.schedules.get(schedule_name or "schedule1")
        schedule.put_on_schedule(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        return subject_consent

    @staticmethod
    def add_unscheduled_appointment(appointment=None):
        creator = UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            facility=appointment.facility,
            timepoint=appointment.timepoint,
        )
        return creator.appointment



from dateutil.relativedelta import relativedelta
from django.test import TestCase, tag
from edc_facility.import_holidays import import_holidays
from edc_protocol import Protocol
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_appointment.constants import INCOMPLETE_APPT, ONTIME_APPT
from edc_appointment.creators import UnscheduledAppointmentCreator
from edc_appointment.forms import AppointmentForm
from edc_appointment.model_mixins import AppointmentWindowError
from edc_appointment.tests.models import SubjectVisit

from ...models import Appointment
from ..helper import Helper
from ..visit_schedule import visit_schedule3


class TestAppointmentWindowPeriod(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule3)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=Protocol().study_open_datetime,
        )

    @staticmethod
    def create_unscheduled(appointment_1030):
        return UnscheduledAppointmentCreator(
            subject_identifier=appointment_1030.subject_identifier,
            visit_schedule_name=appointment_1030.visit_schedule_name,
            schedule_name=appointment_1030.schedule_name,
            visit_code=appointment_1030.visit_code,
            facility=appointment_1030.facility,
        ).appointment

    @staticmethod
    def get_form(appointment, appt_datetime):
        return AppointmentForm(
            data={
                "appt_datetime": appt_datetime,
                "appt_timing": ONTIME_APPT,
                "subject_identifier": appointment.subject_identifier,
                "timepoint_status": appointment.timepoint_status,
                "timepoint": appointment.timepoint,
                "timepoint_datetime": appointment.timepoint_datetime,
                "appt_close_datetime": appointment.timepoint_datetime,
                "facility_name": appointment.facility_name,
                "appt_type": appointment.appt_type,
                "appt_status": appointment.appt_status,
                "appt_reason": appointment.appt_reason,
            },
            instance=appointment,
        )

    @staticmethod
    def update_appt_as_incomplete(appointment):
        appointment.refresh_from_db()
        SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()

    def create_1030_and_1060(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )

        appointment_1000 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1000"
        )
        SubjectVisit.objects.create(
            appointment=appointment_1000,
            report_datetime=appointment_1000.appt_datetime,
            reason=SCHEDULED,
        )
        appointment_1000.appt_status = INCOMPLETE_APPT
        appointment_1000.save()
        appointment_1000.refresh_from_db()

        appointment_1030 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1030"
        )
        SubjectVisit.objects.create(
            appointment=appointment_1030,
            report_datetime=appointment_1030.appt_datetime,
            reason=SCHEDULED,
        )
        appointment_1030.appt_status = INCOMPLETE_APPT
        appointment_1030.save()
        appointment_1030.refresh_from_db()
        return (
            appointment_1030,
            Appointment.objects.get(
                subject_identifier=self.subject_identifier, visit_code="1060"
            ),
        )

    def test_appointments_window_period(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 5)

        appointment_1030 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1030"
        )
        appointment_1060 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1060"
        )
        appointment_1030.appt_datetime = appointment_1060.appt_datetime
        self.assertRaises(AppointmentWindowError, appointment_1030.save)

    @tag("22")
    def test_appointments_window_period_in_form(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )
        appointment_1030 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1030"
        )
        appointment_1060 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1060"
        )
        form = AppointmentForm(
            data={"appt_datetime": appointment_1060.appt_datetime},
            instance=appointment_1030,
        )
        form.is_valid()
        self.assertIn("appt_datetime", form._errors)
        self.assertIn("An appointment already exists", form._errors.get("appt_datetime")[0])

    @tag("1")
    def test_appointments_window_period_boundary_before_next_lower(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        delta = appointment_1060.appt_datetime - appointment_1030.appt_datetime

        appt_datetime_before = (
            appointment_1030.appt_datetime
            + delta
            - appointment_1060.visit_from_schedule.rlower
            - relativedelta(days=1)
        )
        unscheduled_appointment = self.create_unscheduled(appointment_1030)
        form = self.get_form(unscheduled_appointment, appt_datetime_before)
        form.is_valid()
        self.assertNotIn("appt_datetime", form._errors)
        form.save()
        self.update_appt_as_incomplete(unscheduled_appointment)

    @tag("appt")
    def test_appointments_window_period_boundary_on_next_lower(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        delta = appointment_1060.appt_datetime - appointment_1030.appt_datetime

        appt_datetime_on = (
            appointment_1030.appt_datetime
            + delta
            - appointment_1060.visit_from_schedule.rlower
        )
        unscheduled_appointment = self.create_unscheduled(appointment_1030)
        form = self.get_form(unscheduled_appointment, appt_datetime_on)
        form.is_valid()
        self.assertIn("appt_datetime", form._errors)

    @tag("appt")
    def test_appointments_window_period_boundary_after_next_lower(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        delta = appointment_1060.appt_datetime - appointment_1030.appt_datetime

        appt_datetime_after = (
            appointment_1030.appt_datetime
            + delta
            - appointment_1060.visit_from_schedule.rlower
            + relativedelta(days=1)
        )
        unscheduled_appointment = self.create_unscheduled(appointment_1030)
        form = self.get_form(unscheduled_appointment, appt_datetime_after)
        form.is_valid()
        self.assertIn("appt_datetime", form._errors)



from typing import Any

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist

from .choices import DEFAULT_APPT_REASON_CHOICES
from .constants import CANCELLED_APPT, MISSED_APPT, SCHEDULED_APPT, UNSCHEDULED_APPT


def get_appt_reason_choices() -> tuple:
    """Returns a customized tuple of choices otherwise the default"""
    settings_attr = "EDC_APPOINTMENT_APPT_REASON_CHOICES"
    appt_reason_choices = getattr(settings, settings_attr, DEFAULT_APPT_REASON_CHOICES)
    required_keys = [choice[0] for choice in appt_reason_choices]
    for key in [SCHEDULED_APPT, UNSCHEDULED_APPT]:
        if key not in required_keys:
            raise ImproperlyConfigured(
                f"Invalid APPT_REASON_CHOICES. Missing key `{key}`. See {settings_attr}."
            )
    return appt_reason_choices


def cancelled_appointment(instance: Any):
    try:
        cancelled = instance.appt_status == CANCELLED_APPT
    except AttributeError as e:
        if "appt_status" not in str(e):
            raise
    else:
        if (
            cancelled
            and instance.visit_code_sequence > 0
            and "historical" not in instance._meta.label_lower
        ):
            try:
                instance.visit_model_cls().objects.get(appointment=instance)
            except ObjectDoesNotExist:
                instance.delete()


def missed_appointment(instance: Any):
    try:
        missed = instance.appt_timing == MISSED_APPT
    except AttributeError as e:
        if "appt_timing" not in str(e):
            raise
    else:
        if (
            missed
            and instance.visit_code_sequence == 0
            and "historical" not in instance._meta.label_lower
        ):
            try:
                instance.create_missed_visit_from_appointment()
            except AttributeError as e:
                if "create_missed_visit" not in str(e):
                    raise



from dateutil.relativedelta import relativedelta
from edc_visit_schedule import (
    Crf,
    FormsCollection,
    Requisition,
    Schedule,
    Visit,
    VisitSchedule,
)
from edc_visit_schedule.tests.dummy_panel import DummyPanel


class MockPanel(DummyPanel):
    """`requisition_model` is normally set when the lab profile
    is set up.
    """

    def __init__(self, name):
        super().__init__(requisition_model="edc_appointment.subjectrequisition", name=name)


panel_one = MockPanel(name="one")
panel_two = MockPanel(name="two")
panel_three = MockPanel(name="three")
panel_four = MockPanel(name="four")
panel_five = MockPanel(name="five")
panel_six = MockPanel(name="six")

crfs = FormsCollection(
    Crf(show_order=1, model="edc_metadata.crfone", required=True),
    Crf(show_order=2, model="edc_metadata.crftwo", required=True),
    Crf(show_order=3, model="edc_metadata.crfthree", required=True),
    Crf(show_order=4, model="edc_metadata.crffour", required=True),
    Crf(show_order=5, model="edc_metadata.crffive", required=True),
)

crfs_missed = FormsCollection(
    Crf(show_order=10, model="edc_metadata.subjectvisitmissed"),
    name="missed",
)

requisitions = FormsCollection(
    Requisition(show_order=10, panel=panel_one, required=True, additional=False),
    Requisition(show_order=20, panel=panel_two, required=True, additional=False),
    Requisition(show_order=30, panel=panel_three, required=True, additional=False),
    Requisition(show_order=40, panel=panel_four, required=True, additional=False),
    Requisition(show_order=50, panel=panel_five, required=True, additional=False),
    Requisition(show_order=60, panel=panel_six, required=True, additional=False),
)


crfs_unscheduled = FormsCollection(
    Crf(show_order=1, model="edc_metadata.crfone", required=True),
    Crf(show_order=3, model="edc_metadata.crfthree", required=True),
    Crf(show_order=5, model="edc_metadata.crffive", required=True),
)


visit_schedule1 = VisitSchedule(
    name="visit_schedule1",
    offstudy_model="edc_appointment.subjectoffstudy",
    death_report_model="edc_appointment.deathreport",
    locator_model="edc_appointment.subjectlocator",
)

visit_schedule2 = VisitSchedule(
    name="visit_schedule2",
    offstudy_model="edc_appointment.subjectoffstudy2",
    death_report_model="edc_appointment.deathreport",
    locator_model="edc_appointment.subjectlocator",
)

visit_schedule3 = VisitSchedule(
    name="visit_schedule3",
    offstudy_model="edc_appointment.subjectoffstudy2",
    death_report_model="edc_appointment.deathreport",
    locator_model="edc_appointment.subjectlocator",
)

schedule1 = Schedule(
    name="schedule1",
    onschedule_model="edc_appointment.onscheduleone",
    offschedule_model="edc_appointment.offscheduleone",
    appointment_model="edc_appointment.appointment",
    consent_model="edc_appointment.subjectconsent",
)

schedule2 = Schedule(
    name="schedule2",
    onschedule_model="edc_appointment.onscheduletwo",
    offschedule_model="edc_appointment.offscheduletwo",
    appointment_model="edc_appointment.appointment",
    consent_model="edc_appointment.subjectconsent",
)


schedule3 = Schedule(
    name="three_monthly_schedule",
    onschedule_model="edc_appointment.onschedulethree",
    offschedule_model="edc_appointment.offschedulethree",
    appointment_model="edc_appointment.appointment",
    consent_model="edc_appointment.subjectconsent",
)

visits = []
for index in range(0, 4):
    visits.append(
        Visit(
            code=f"{1 if index == 0 else index + 1}000",
            title=f"Day {1 if index == 0 else index + 1}",
            timepoint=index,
            rbase=relativedelta(days=7 * index),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            requisitions=requisitions,
            crfs=crfs,
            crfs_missed=crfs_missed,
            requisitions_unscheduled=requisitions,
            crfs_unscheduled=crfs_unscheduled,
            allow_unscheduled=True,
            facility_name="5-day-clinic",
        )
    )
for visit in visits:
    schedule1.add_visit(visit)

visits = []
for index in range(4, 8):
    visits.append(
        Visit(
            code=f"{1 if index == 0 else index + 1}000",
            title=f"Day {1 if index == 0 else index + 1}",
            timepoint=index,
            rbase=relativedelta(days=7 * index),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            requisitions=requisitions,
            crfs=crfs,
            crfs_missed=crfs_missed,
            facility_name="7-day-clinic",
        )
    )
for visit in visits:
    schedule2.add_visit(visit)

visits = []
visits.append(
    Visit(
        code="1000",
        title="Baseline",
        timepoint=0,
        rbase=relativedelta(days=0),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=0),
        requisitions=requisitions,
        crfs=crfs,
        crfs_missed=crfs_missed,
        facility_name="7-day-clinic",
    )
)

for index, visit_code in [(3, "1030"), (6, "1060"), (9, "1090"), (12, "1120")]:
    visits.append(
        Visit(
            code=visit_code,
            title=f"Month {index}",
            timepoint=index,
            rbase=relativedelta(months=index),
            rlower=relativedelta(days=14),
            rupper=relativedelta(days=45),
            requisitions=requisitions,
            crfs=crfs,
            crfs_missed=crfs_missed,
            facility_name="7-day-clinic",
            allow_unscheduled=True,
        )
    )
for visit in visits:
    schedule3.add_visit(visit)

visit_schedule1.add_schedule(schedule1)
visit_schedule2.add_schedule(schedule2)
visit_schedule3.add_schedule(schedule3)



from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from ..constants import (
    CANCELLED_APPT,
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    NEW_APPT,
    UNSCHEDULED_APPT,
)
from .appointment_creator import AppointmentCreator


class UnscheduledAppointmentError(Exception):
    pass


class UnscheduledAppointmentNotAllowed(Exception):
    pass


class InvalidParentAppointmentStatusError(Exception):
    pass


class InvalidParentAppointmentMissingVisitError(Exception):
    pass


class AppointmentInProgressError(Exception):
    pass


class UnscheduledAppointmentCreator:

    appointment_creator_cls = AppointmentCreator

    def __init__(
        self,
        subject_identifier=None,
        appt_datetime=None,
        visit_schedule_name=None,
        schedule_name=None,
        visit_code=None,
        facility=None,
        **kwargs,
    ):
        self._parent_appointment = None
        self.appointment = None
        self.subject_identifier = subject_identifier
        self.visit_schedule_name = visit_schedule_name
        self.schedule_name = schedule_name
        self.visit_code = visit_code
        self.facility = facility
        self.visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name)
        self.schedule = self.visit_schedule.schedules.get(schedule_name)
        self.appointment_model_cls = self.schedule.appointment_model_cls
        visit = self.visit_schedule.schedules.get(schedule_name).visits.get(visit_code)
        if not visit:
            raise UnscheduledAppointmentError(
                "Invalid visit. Got None using {"
                f"visit_schedule_name='{visit_schedule_name}',"
                f"schedule_name='{schedule_name}',"
                f"visit_code='{visit_code}'" + "}"
            )
        if visit.allow_unscheduled:
            # force lookup and parent_appointment exceptions
            self.parent_appointment
            # do not allow if any appointments are IN_PROGRESS
            try:
                obj = self.appointment_model_cls.objects.get(
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=self.visit_schedule_name,
                    schedule_name=self.schedule_name,
                    appt_status=IN_PROGRESS_APPT,
                )
            except MultipleObjectsReturned as e:
                raise UnscheduledAppointmentError(e)
            except ObjectDoesNotExist:
                pass
            else:
                raise AppointmentInProgressError(
                    f"Not allowed. Appointment {obj.visit_code}."
                    f"{obj.visit_code_sequence} is in progress."
                )

            # don't allow if next appointment is already started.
            next_by_timepoint = self.parent_appointment.next_by_timepoint
            if next_by_timepoint:
                if next_by_timepoint.appt_status not in [NEW_APPT, CANCELLED_APPT]:
                    raise UnscheduledAppointmentError(
                        f"Not allowed. Visit {next_by_timepoint.visit_code} has "
                        "already been started."
                    )
            appointment_creator = self.appointment_creator_cls(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit=visit,
                suggested_datetime=(appt_datetime or self.parent_appointment.appt_datetime),
                timepoint=self.parent_appointment.timepoint,
                timepoint_datetime=self.parent_appointment.timepoint_datetime,
                visit_code_sequence=self.parent_appointment.next_visit_code_sequence,
                facility=self.facility,
                appt_status=IN_PROGRESS_APPT,
                appt_reason=UNSCHEDULED_APPT,
            )
            self.appointment = appointment_creator.appointment
        else:
            raise UnscheduledAppointmentNotAllowed(
                f"Not allowed. Visit {visit_code} is not configured for "
                "unscheduled appointments."
            )

    @property
    def parent_appointment(self):
        if not self._parent_appointment:
            options = dict(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence=0,
            )
            self._parent_appointment = self.appointment_model_cls.objects.get(**options)
            try:
                self._parent_appointment.visit
            except ObjectDoesNotExist:
                raise InvalidParentAppointmentMissingVisitError(
                    f"Unable to create unscheduled appointment. An unscheduled "
                    f"appointment cannot be created if the parent appointment "
                    f"visit form has not been completed. "
                    f"Got appointment '{self.visit_code}.0'."
                )
            else:
                if self._parent_appointment.appt_status not in [
                    COMPLETE_APPT,
                    INCOMPLETE_APPT,
                ]:
                    raise InvalidParentAppointmentStatusError(
                        f"Unable to create unscheduled appointment. An unscheduled "
                        f"appointment cannot be created if the parent appointment "
                        f"is 'new' or 'in progress'. Got appointment "
                        f"'{self.visit_code}' is "
                        f"{self._parent_appointment.get_appt_status_display().lower()}."
                    )

        return self._parent_appointment



from django.core.exceptions import ObjectDoesNotExist
from edc_visit_schedule import site_visit_schedules
from import_export.fields import Field
from import_export.resources import ModelResource

from .models import Appointment


class AppointmentResource(ModelResource):

    offstudy_datetime = Field(column_name="offstudy_datetime")

    def dehydrate_offstudy(self, obj):
        visit_schedule = site_visit_schedules.get_visit_schedule(
            visit_schedule_name=obj.visit_schedule_name
        )
        schedule = visit_schedule.schedules.get(obj.schedule_name)
        try:
            off_schedule_obj = schedule.offschedule_model_cls.objects.get(
                subject_identifier=obj.subject_identifier,
                report_datetime__lte=obj.appt_datetime,
            )
        except ObjectDoesNotExist:
            return None
        return off_schedule_obj.offschedule_datetime

    class Meta:
        model = Appointment



from django.db import models
from edc_visit_schedule.schedule.window import (
    ScheduledVisitWindowError,
    UnScheduledVisitWindowError,
)
from edc_visit_schedule.utils import is_baseline

from ..stubs import AppointmentModelStub


class AppointmentWindowError(Exception):
    pass


class WindowPeriodModelMixin(models.Model):
    """A model mixin declared with the Appointment model to managed
    window period calculations for appt_datetime.
    """

    window_period_checks_enabled = True

    def save(self: AppointmentModelStub, *args, **kwargs) -> None:
        if self.id and self.appt_datetime and self.timepoint_datetime:
            self.raise_on_not_datetime_in_window()
        super().save(*args, **kwargs)  # type:ignore

    def raise_on_not_datetime_in_window(self: AppointmentModelStub):
        if not self.is_baseline_appt:
            baseline_timepoint_datetime = self.__class__.objects.first_appointment(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
            ).timepoint_datetime
            try:
                self.schedule.datetime_in_window(
                    dt=self.appt_datetime,
                    timepoint_datetime=self.timepoint_datetime,
                    visit_code=self.visit_code,
                    visit_code_sequence=self.visit_code_sequence,
                    baseline_timepoint_datetime=baseline_timepoint_datetime,
                )
            except ScheduledVisitWindowError as e:
                msg = str(e)
                msg.replace("Invalid datetime", "Invalid appointment datetime")
                msg = f"{msg} Perhaps catch this in the form."
                raise AppointmentWindowError(msg)
            except UnScheduledVisitWindowError as e:
                msg = str(e)
                msg.replace("Invalid datetime", "Invalid appointment datetime")
                msg = f"{msg} Perhaps catch this in the form."
                raise AppointmentWindowError(msg)

    @property
    def is_baseline_appt(self: AppointmentModelStub) -> bool:
        return is_baseline(instance=self)

    class Meta:
        abstract = True



from datetime import datetime
from logging import warning
from typing import Any
from zoneinfo import ZoneInfo

from arrow.arrow import Arrow
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.urls import reverse
from django.utils.html import format_html
from edc_consent import NotConsentedError
from edc_consent.requires_consent import RequiresConsent
from edc_consent.site_consents import SiteConsentError, site_consents
from edc_constants.constants import INCOMPLETE
from edc_form_validators import INVALID_ERROR
from edc_form_validators.form_validator import FormValidator
from edc_metadata.metadata_helper_mixin import MetaDataHelperMixin
from edc_registration import get_registered_subject_model_cls
from edc_utils import formatted_datetime, get_utcnow
from edc_visit_schedule import site_visit_schedules
from edc_visit_schedule.subject_schedule import NotOnScheduleError
from edc_visit_schedule.utils import get_onschedule_model_instance
from edc_visit_tracking.reason_updater import (
    SubjectVisitReasonUpdater,
    SubjectVisitReasonUpdaterBaselineError,
    SubjectVisitReasonUpdaterCrfsExistsError,
    SubjectVisitReasonUpdaterMissedVisitNotAllowed,
    SubjectVisitReasonUpdaterRequisitionsExistsError,
)

from ..constants import (
    CANCELLED_APPT,
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    NEW_APPT,
    UNSCHEDULED_APPT,
)
from .window_period_form_validator_mixin import WindowPeriodFormValidatorMixin

INVALID_APPT_DATE = "invalid_appt_datetime"
INVALID_APPT_STATUS = "invalid_appt_status"
INVALID_APPT_REASON = "invalid_appt_reason"
INVALID_PREVIOUS_VISIT_MISSING = "previous_visit_missing"
INVALID_MISSED_APPT_NOT_ALLOWED = "invalid_missed_appt_not_allowed"
INVALID_MISSED_APPT_NOT_ALLOWED_AT_BASELINE = "invalid_missed_appt_not_allowed_at_baseline"
INVALID_APPT_TIMING_CRFS_EXIST = "invalid_appt_timing_crfs_exist"
INVALID_APPT_TIMING_REQUISITIONS_EXIST = "invalid_appt_timing_requisitions_exist"


class AppointmentFormValidator(
    MetaDataHelperMixin, WindowPeriodFormValidatorMixin, FormValidator
):
    """Note, the appointment is only changed, never added,
    through this form.
    """

    appointment_model = "edc_appointment.appointment"

    def clean(self: Any):
        # TODO: do not allow a missed appt (in window) to be followed by an unscheduled appt
        #  that is also within window.
        self.validate_visit_report_sequence()
        self.validate_appt_sequence()
        self.validate_timepoint()
        self.validate_appt_datetime_unique()
        self.validate_appt_datetime_not_before_consent_datetime()
        self.validate_appt_datetime_not_after_next_appt_datetime()
        self.validate_not_future_appt_datetime()
        self.validate_appt_datetime_in_window_period(
            self.instance,
            self.cleaned_data.get("appt_datetime"),
            "appt_datetime",
        )
        self.validate_subject_on_schedule()
        self.validate_appt_reason()
        self.validate_appt_new_or_cancelled()
        self.validate_appt_inprogress_or_incomplete()
        self.validate_appt_inprogress()
        self.validate_appt_new_or_complete()

        self.validate_facility_name()

        self.validate_appointment_timing()
        rappt_datetime = Arrow.fromdatetime(
            self.cleaned_data.get("appt_datetime"), self.instance.appt_datetime.tzinfo
        )
        if rappt_datetime.to("UTC").date() != self.instance.appt_datetime:
            self.update_subject_visit_document_status(INCOMPLETE)

    @property
    def appointment_model_cls(self) -> Any:
        return django_apps.get_model(self.appointment_model)

    @property
    def required_additional_forms_exist(self) -> bool:
        """Returns True if any additional required forms are
        yet to be keyed.
        """
        return False

    def validate_visit_report_sequence(self: Any) -> bool:
        """Enforce visit report sequence."""
        try:
            self.instance.visit
        except (ObjectDoesNotExist, AttributeError):
            if self.cleaned_data.get("appt_status") == IN_PROGRESS_APPT and self.instance:
                try:
                    self.instance.visit
                except ObjectDoesNotExist:
                    previous_appt = self.instance.get_previous(include_interim=True)
                    if previous_appt and previous_appt.appt_status != CANCELLED_APPT:
                        try:
                            previous_appt.visit
                        except ObjectDoesNotExist:
                            self.raise_validation_error(
                                message=(
                                    "A previous appointment requires a visit report. "
                                    f"Update appointment {previous_appt.visit_code}."
                                    f"{previous_appt.visit_code_sequence} first."
                                ),
                                error_code=INVALID_PREVIOUS_VISIT_MISSING,
                            )
        return True

    def validate_appt_sequence(self: Any) -> bool:
        """Enforce appointment and visit entry sequence.

        1. Check if previous appointment has a visit report
        2. If not check which previous appointment, if any,
        has a completed visit report
        3. If none, is this the first appointment?

        """
        if self.cleaned_data.get("appt_status") in [
            IN_PROGRESS_APPT,
            INCOMPLETE_APPT,
            COMPLETE_APPT,
        ]:
            try:
                self.instance.previous.visit
            except ObjectDoesNotExist:
                first_new_appt = (
                    self.appointment_model_cls.objects.filter(
                        subject_identifier=self.instance.subject_identifier,
                        visit_schedule_name=self.instance.visit_schedule_name,
                        schedule_name=self.instance.schedule_name,
                        appt_status=NEW_APPT,
                    )
                    .order_by("timepoint", "visit_code_sequence")
                    .first()
                )
                if first_new_appt:
                    self.raise_validation_error(
                        {
                            "__all__": (
                                "A previous appointment requires updating. "
                                f"Update appointment for {first_new_appt.visit_code} first."
                            )
                        },
                        INVALID_ERROR,
                    )
            except AttributeError:
                pass
        return True

    def validate_timepoint(self):
        try:
            timepoint = self.instance.timepoint
        except AttributeError:
            pass
        else:
            visit_schedule = site_visit_schedules.get_visit_schedule(
                self.instance.visit_schedule_name
            )
            schedule = visit_schedule.schedules.get(self.instance.schedule_name)
            visit = schedule.visits.get(self.instance.visit_code)
            if visit and timepoint != visit.timepoint:
                self.raise_validation_error(
                    f"Invalid timepoint. Expected {visit.timepoint} "
                    f"for visit_code={visit.visit_code}. Got {timepoint}"
                )

    def validate_not_future_appt_datetime(self: Any) -> None:
        appt_datetime = self.cleaned_data.get("appt_datetime")
        if appt_datetime and appt_datetime != NEW_APPT:
            rappt_datetime = Arrow.fromdatetime(appt_datetime, appt_datetime.tzinfo)
            if rappt_datetime.to("UTC").date() > get_utcnow().date():
                self.raise_validation_error(
                    {"appt_datetime": "Cannot be a future date."}, INVALID_APPT_DATE
                )

    def validate_appt_datetime_unique(self):
        """Assert one visit report per day"""
        if self.cleaned_data.get("appt_datetime"):
            tz = ZoneInfo("utc")
            appt_date = (
                Arrow.fromdatetime(self.cleaned_data.get("appt_datetime")).to(tz).date()
            )
            try:
                obj = self.instance.__class__.objects.get(appt_datetime__date=appt_date)
            except ObjectDoesNotExist:
                pass
            except MultipleObjectsReturned:
                raise self.raise_validation_error(
                    {"appt_datetime": "An appointment already exists for this date"},
                    INVALID_ERROR,
                )
            else:
                if self.instance and obj.id != self.instance.id:
                    raise self.raise_validation_error(
                        {
                            "appt_datetime": "An appointment already exists for this date. "
                            f"See {obj.visit_code}.{obj.visit_code_sequence}"
                        },
                        INVALID_ERROR,
                    )

    def validate_appt_datetime_not_before_consent_datetime(self: Any) -> None:
        if (
            "edc_consent" not in settings.INSTALLED_APPS
            and "edc_consent.apps.AppConfig" not in settings.INSTALLED_APPS
        ):
            warning(
                "Skipping consent_datetime form validation. "
                "`edc_consent` not in `INSTALLED_APPS`"
            )
        else:
            appt_datetime = self.cleaned_data.get("appt_datetime")
            if appt_datetime and appt_datetime != NEW_APPT:
                rappt_datetime = Arrow.fromdatetime(appt_datetime, appt_datetime.tzinfo)
                consent_datetime = self.consent_datetime_or_raise(
                    rappt_datetime.to("UTC").datetime
                )
                if rappt_datetime.to("UTC").date() < consent_datetime.date():
                    rconsent_datetime = Arrow.fromdatetime(
                        consent_datetime, appt_datetime.tzinfo
                    )
                    formatted_date = formatted_datetime(
                        rconsent_datetime.datetime, format_as_date=True
                    )
                    self.raise_validation_error(
                        {
                            "appt_datetime": (
                                (
                                    "Invalid. Cannot be before consent date. "
                                    f"Got consented on {formatted_date}"
                                )
                            )
                        },
                        INVALID_APPT_DATE,
                    )

    def validate_appointment_timing(self) -> None:
        """Updates the subject visit report (if it exists) as missed or scheduled
        based on appt_timing OR raises"""
        subject_visit_reason_updater = SubjectVisitReasonUpdater(
            appointment=self.instance,
            appt_timing=self.cleaned_data.get("appt_timing"),
            appt_reason=self.cleaned_data.get("appt_reason"),
        )
        try:
            subject_visit_reason_updater.update_or_raise()
        except SubjectVisitReasonUpdaterBaselineError as e:
            self.raise_validation_error(
                {"appt_timing": str(e)}, INVALID_MISSED_APPT_NOT_ALLOWED_AT_BASELINE
            )
        except SubjectVisitReasonUpdaterMissedVisitNotAllowed as e:
            self.raise_validation_error(
                {"appt_timing": str(e)}, INVALID_MISSED_APPT_NOT_ALLOWED
            )

        except SubjectVisitReasonUpdaterCrfsExistsError as e:
            self.raise_validation_error(
                {"appt_timing": str(e)}, INVALID_APPT_TIMING_CRFS_EXIST
            )

        except SubjectVisitReasonUpdaterRequisitionsExistsError as e:
            self.raise_validation_error(
                {"appt_timing": str(e)}, INVALID_APPT_TIMING_REQUISITIONS_EXIST
            )

    def validate_appt_datetime_not_after_next_appt_datetime(self: Any) -> None:
        appt_datetime = self.cleaned_data.get("appt_datetime")
        if appt_datetime and appt_datetime != NEW_APPT:
            rappt_datetime = Arrow.fromdatetime(appt_datetime, appt_datetime.tzinfo)
            if self.instance.next:
                if rappt_datetime.to("UTC").date() > self.instance.next.appt_datetime.date():
                    formatted_date = formatted_datetime(
                        self.instance.next.appt_datetime, format_as_date=True
                    )
                    self.raise_validation_error(
                        {
                            "appt_datetime": (
                                "Cannot be after next scheduled appointment. "
                                f"Next appointment is on {formatted_date}"
                            )
                        },
                        INVALID_APPT_DATE,
                    )

    def validate_appt_new_or_cancelled(self: Any) -> None:
        """Don't allow new or cancelled if form data exists
        and don't allow cancelled if visit_code_sequence == 0.
        """
        appt_status = self.cleaned_data.get("appt_status")
        if self.instance.visit_code_sequence == 0 and appt_status == CANCELLED_APPT:
            self.raise_validation_error(
                {"appt_status": "Invalid. A scheduled appointment may not be cancelled."},
                INVALID_APPT_STATUS,
            )
        elif appt_status in [NEW_APPT, CANCELLED_APPT] and self.crf_metadata_keyed_exists:
            self.raise_validation_error(
                {"appt_status": "Invalid. CRF data has already been entered."},
                INVALID_APPT_STATUS,
            )
        elif (
            appt_status in [NEW_APPT, CANCELLED_APPT]
            and self.requisition_metadata_keyed_exists
        ):
            self.raise_validation_error(
                {"appt_status": "Invalid. requisition data has already been entered."},
                INVALID_APPT_STATUS,
            )

    def validate_appt_inprogress_or_incomplete(self: Any) -> None:
        appt_status = self.cleaned_data.get("appt_status")
        if appt_status == CANCELLED_APPT and self.crf_metadata_keyed_exists:
            self.raise_validation_error(
                {"appt_status": format_html("Invalid. Some CRFs have already been keyed")},
                INVALID_APPT_STATUS,
            )
        elif appt_status == CANCELLED_APPT and self.requisition_metadata_keyed_exists:
            self.raise_validation_error(
                {
                    "appt_status": format_html(
                        "Invalid. Some requisitions have already been keyed"
                    )
                },
                INVALID_APPT_STATUS,
            )

        elif (
            appt_status not in [INCOMPLETE_APPT, IN_PROGRESS_APPT, CANCELLED_APPT]
            and self.crf_metadata_required_exists
        ):
            url = self.changelist_url("crfmetadata")
            self.raise_validation_error(
                {
                    "appt_status": format_html(
                        f'Invalid. Not all <a href="{url}">required CRFs</a> have been keyed'
                    )
                },
                INVALID_APPT_STATUS,
            )
        elif (
            appt_status not in [INCOMPLETE_APPT, IN_PROGRESS_APPT, CANCELLED_APPT]
            and self.requisition_metadata_required_exists
        ):
            url = self.changelist_url("requisitionmetadata")
            self.raise_validation_error(
                {
                    "appt_status": format_html(
                        f'Invalid. Not all <a href="{url}">'
                        "required requisitions</a> have been keyed"
                    )
                },
                INVALID_APPT_STATUS,
            )

    def validate_appt_inprogress(self: Any) -> None:
        appt_status = self.cleaned_data.get("appt_status")
        if appt_status == IN_PROGRESS_APPT and self.appointment_in_progress_exists:
            self.raise_validation_error(
                {
                    "appt_status": (
                        "Invalid. Another appointment in this schedule is in progress."
                    )
                },
                INVALID_APPT_STATUS,
            )

    def validate_appt_new_or_complete(self: Any) -> None:
        appt_status = self.cleaned_data.get("appt_status")
        if (
            appt_status not in [COMPLETE_APPT, NEW_APPT]
            and self.crf_metadata_exists
            and self.requisition_metadata_exists
            and not self.crf_metadata_required_exists
            and not self.requisition_metadata_required_exists
            and not self.required_additional_forms_exist
        ):
            if not self.crf_metadata_required_exists:
                self.raise_validation_error(
                    {"appt_status": "Invalid. All required CRFs have been keyed"},
                    INVALID_APPT_STATUS,
                )
            elif not self.requisition_metadata_required_exists:
                self.raise_validation_error(
                    {"appt_status": "Invalid. All required requisitions have been keyed"},
                    INVALID_APPT_STATUS,
                )
            elif not self.required_additional_forms_exist:
                self.raise_validation_error(
                    {
                        "appt_status": (
                            "Invalid. All required 'additional' forms have been keyed"
                        )
                    },
                    INVALID_APPT_STATUS,
                )

    @property
    def appointment_in_progress_exists(self: Any) -> None:
        """Returns True if another appointment in this schedule
        is currently set to "in_progress".
        """
        return (
            self.appointment_model_cls.objects.filter(
                subject_identifier=self.instance.subject_identifier,
                visit_schedule_name=self.instance.visit_schedule_name,
                schedule_name=self.instance.schedule_name,
                appt_status=IN_PROGRESS_APPT,
            )
            .exclude(id=self.instance.id)
            .exists()
        )

    def validate_facility_name(self: Any) -> None:
        """Raises if facility_name not found in edc_facility.AppConfig."""
        if self.cleaned_data.get("facility_name"):
            app_config = django_apps.get_app_config("edc_facility")
            if self.cleaned_data.get("facility_name") not in app_config.facilities:
                self.raise_validation_error(
                    {"__all__": f"Facility '{self.facility_name}' does not exist."},
                    INVALID_ERROR,
                )

    def validate_appt_reason(self: Any) -> None:
        """Raises if visit_code_sequence is not 0 and appt_reason
        is not UNSCHEDULED_APPT.
        """
        appt_reason = self.cleaned_data.get("appt_reason")
        appt_status = self.cleaned_data.get("appt_status")
        if (
            appt_reason
            and self.instance.visit_code_sequence
            and appt_reason != UNSCHEDULED_APPT
        ):
            self.raise_validation_error(
                {"appt_reason": f"Expected {UNSCHEDULED_APPT.title()}"}, INVALID_APPT_REASON
            )
        elif (
            appt_reason
            and not self.instance.visit_code_sequence
            and appt_reason == UNSCHEDULED_APPT
        ):
            self.raise_validation_error(
                {"appt_reason": f"Cannot be {UNSCHEDULED_APPT.title()}"}, INVALID_APPT_REASON
            )
        elif (
            appt_status
            and not self.instance.visit_code_sequence
            and appt_status == CANCELLED_APPT
        ):
            self.raise_validation_error(
                {"appt_status": "Invalid. A scheduled appointment cannot be cancelled"},
                INVALID_APPT_STATUS,
            )

    def changelist_url(self: Any, model_name: str) -> Any:
        """Returns the model's changelist url with filter querystring"""
        url = reverse(f"edc_metadata_admin:edc_metadata_{model_name}_changelist")
        url = (
            f"{url}?q={self.instance.subject_identifier}"
            f"&visit_code={self.instance.visit_code}"
            f"&visit_code_sequence={self.instance.visit_code_sequence}"
        )
        return url

    def consent_datetime_or_raise(self: Any, appt_datetime: datetime) -> datetime:
        consent_model = site_visit_schedules.get_consent_model(
            visit_schedule_name=self.instance.visit_schedule_name,
            schedule_name=self.instance.schedule_name,
        )
        try:
            RequiresConsent(
                model=self.instance._meta.label_lower,
                subject_identifier=self.instance.subject_identifier,
                report_datetime=appt_datetime,
                consent_model=consent_model,
            )
        except SiteConsentError:
            self.raise_validation_error(
                {
                    "appt_datetime": (
                        "Date does not fall within any configured consent period."
                        f"Expected one of {site_consents.consents}. "
                        "See the EDC configuration."
                    )
                },
                INVALID_APPT_DATE,
            )
        except NotConsentedError as e:
            self.raise_validation_error(
                {"appt_datetime": str(e)},
                INVALID_APPT_DATE,
            )
        registered_subject = get_registered_subject_model_cls().objects.get(
            subject_identifier=self.instance.subject_identifier
        )
        return registered_subject.consent_datetime

    def validate_subject_on_schedule(self: Any) -> None:
        appt_datetime = self.cleaned_data.get("appt_datetime")
        subject_identifier = self.instance.subject_identifier
        onschedule_model = site_visit_schedules.get_onschedule_model(
            visit_schedule_name=self.instance.visit_schedule_name,
            schedule_name=self.instance.schedule_name,
        )
        qs = django_apps.get_model(onschedule_model).objects.filter(
            subject_identifier=subject_identifier,
        )

        try:
            get_onschedule_model_instance(
                subject_identifier=subject_identifier,
                visit_schedule_name=self.instance.visit_schedule_name,
                schedule_name=self.instance.schedule_name,
                reference_datetime=appt_datetime,
            )
        except NotOnScheduleError:
            self.raise_validation_error(
                (
                    "Subject is not on a schedule for the given date and time. "
                    f"Expected one of {[str(obj) for obj in qs.all()]}. "
                    "Check the appointment date and/or time"
                ),
                INVALID_APPT_DATE,
            )

    def update_subject_visit_document_status(self: Any, document_status: str) -> None:
        """Updates the subject visit document status.

        Changing document status to incomplete forces the
        user to edit the subject visit on the dashboard
        before proceeding to the CRFs.
        """
        try:
            subject_visit = getattr(self.instance, self.instance.related_visit_model_attr())
        except ObjectDoesNotExist:
            pass
        else:
            if subject_visit.document_status != document_status:
                subject_visit.document_status = document_status
                subject_visit.save(update_fields=["document_status"])



from django import forms
from edc_form_validators import FormValidatorMixin
from edc_sites.forms import SiteModelFormMixin

from ..form_validators import AppointmentFormValidator
from ..models import Appointment
from ..utils import get_appt_reason_choices

appt_reason_fld = Appointment._meta.get_field("appt_reason")


class AppointmentForm(SiteModelFormMixin, FormValidatorMixin, forms.ModelForm):
    """Note, the appointment is only changed, never added,
    through this form.
    """

    form_validator_cls = AppointmentFormValidator

    class Meta:
        model = Appointment
        fields = "__all__"
        widgets = {
            "appt_reason": forms.RadioSelect(
                attrs={
                    "label": appt_reason_fld.verbose_name,
                    "required": True,
                    "help_text": appt_reason_fld.help_text,
                },
                choices=get_appt_reason_choices(),
            ),
        }



from edc_auth.site_auths import site_auths

from .auth_objects import APPOINTMENT, APPOINTMENT_EXPORT, APPOINTMENT_VIEW

site_auths.add_group(
    "edc_appointment.view_appointment",
    "edc_appointment.view_historicalappointment",
    name=APPOINTMENT_VIEW,
)

site_auths.add_group(
    "edc_appointment.add_appointment",
    "edc_appointment.change_appointment",
    "edc_appointment.view_appointment",
    "edc_appointment.view_historicalappointment",
    name=APPOINTMENT,
)

site_auths.add_group(
    "edc_appointment.export_appointment",
    name=APPOINTMENT_EXPORT,
)



# Generated by Django 3.0.6 on 2020-07-29 20:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_appointment", "0021_auto_20191024_1000"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="appointment",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("timepoint", "visit_code_sequence"),
            },
        ),
    ]



from django.conf import settings
from django.views.generic.base import TemplateView
from edc_dashboard.view_mixins import EdcViewMixin


class HomeView(EdcViewMixin, TemplateView):

    template_name = f"edc_appointment/bootstrap{settings.EDC_BOOTSTRAP}/home.html"



from typing import Any

from edc_metadata.metadata_helper_mixin import MetaDataHelperMixin

from .constants import COMPLETE_APPT, IN_PROGRESS_APPT, INCOMPLETE_APPT


class AppointmentStatusUpdaterError(Exception):
    pass


class AppointmentStatusUpdater(MetaDataHelperMixin):

    instance_attr = "appointment"

    def __init__(self, appointment: Any):
        self.appointment = appointment
        if not getattr(self.appointment, "id", None):
            raise AppointmentStatusUpdaterError(
                "Appointment instance must exist. Got `id` is None"
            )
        if self.appointment.appt_status != IN_PROGRESS_APPT:
            self.update_others_from_in_progress()
            self.appointment.appt_status = IN_PROGRESS_APPT
            self.appointment.save_base(update_fields=["appt_status"])

    def update_others_from_in_progress(self):
        opts = dict(
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            appt_status=IN_PROGRESS_APPT,
        )
        for appointment in self.appointment.__class__.objects.filter(**opts).exclude(
            id=self.appointment.id
        ):
            if self.crf_metadata_required_exists or self.requisition_metadata_required_exists:
                appointment.appt_status = INCOMPLETE_APPT
            else:
                appointment.appt_status = COMPLETE_APPT
            appointment.save_base(update_fields=["appt_status"])



# Generated by Django 2.0 on 2018-01-02 21:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("edc_appointment", "0012_auto_20171128_2153")]

    operations = [migrations.DeleteModel(name="Holiday")]
